"""Module defining MetaFrames, used for creating MetaFrames from DataFrames."""

from __future__ import annotations

import json
import pathlib
from datetime import datetime
from importlib.metadata import version
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Union, no_type_check
from warnings import warn

import numpy as np
import polars as pl

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # type: ignore  # noqa

from tqdm import tqdm

from metasyn.config import MetaConfig
from metasyn.file import BaseFileInterface, file_interface_from_dict
from metasyn.privacy import BasePrivacy, get_privacy
from metasyn.util import set_global_seeds
from metasyn.validation import validate_gmf_dict
from metasyn.var import MetaVar
from metasyn.varspec import VarSpec


class MetaFrame:
    """Container for statistical metadata describing a dataset.

    This class is used to fit a MetaFrame to a Polars DataFrame, serialize and
    save the MetaFrame to a file, read a MetaFrame from a file, and create
    a synthetic Polars DataFrame.

    A MetaFrame represents a metadata frame, which is a structure that holds
    statistical metadata about a dataset. The data contained in a MetaFrame
    follows the Generative Metadata Format (GMF).
    The metadata is contained in a collection of MetaVar objects,
    with each MetaVar representing a column (variable).

    A MetaFrame can easily be created using the
    ``fit_dataframe`` method, which takes a Polars DataFrame and fits a
    MetaFrame to it.

    Parameters
    ----------
    meta_vars:
        List of variables representing columns in a DataFrame.
    n_rows:
        Number of rows in the original DataFrame.
    privacy_package:
        Package that supplies the distributions.
    """

    def __init__(
        self,
        meta_vars: List[MetaVar],
        n_rows: Optional[int] = None,
        file_format: Union[None, BaseFileInterface, dict[str, Any]] = None,
    ):
        self.meta_vars = meta_vars
        self.n_rows = n_rows
        self._file_format: Union[None, dict[str, Any]]
        self.file_format = file_format  # type: ignore

    @property
    def n_columns(self) -> int:
        """int: Number of columns of the original dataframe."""
        return len(self.meta_vars)

    @classmethod
    def fit_dataframe(  # noqa: PLR0912
        cls,
        df: Optional[pl.DataFrame],
        var_specs: Optional[Union[list[VarSpec]]] = None,
        plugins: Optional[list[str]] = None,
        privacy: Optional[Union[BasePrivacy, dict]] = None,
        n_rows: Optional[int] = None,
        progress_bar: bool = True,
        config: Optional[Union[pathlib.Path, str, MetaConfig]] = None,
        file_format: Union[dict[str, Any], BaseFileInterface, None] = None,
    ):
        """Create a metasyn object from a polars (or pandas) dataframe.

        The Polars dataframe should be formatted already with the correct
        datatypes, such as pl.Categorical (or the pandas equivalent).

        Parameters
        ----------
        df:
            Polars dataframe with the correct column dtypes.
        var_specs:
            Specifications for each column/variable. These specifications are supplied as
            a list of VarSpec instances (one for each column). Alternatively, the
            specifications can be entered as a path to a .toml file. For more information
            on this approach, see the MetaConfig class or the examples in the documentation.
            By default var_specs is None, which will use the default settings for each column.
        plugins:
            Plugins to use when fitting distributions to variables.
            Can be a list of strings or unspecified. This will overwrite the defaults if they
            were specified in the varspecs (but not the specifications per column).
        privacy:
            Privacy level to use by default. This will overwrite the defaults if they were
            specified in the varspecs (but not the specifications per column).
        n_rows:
            Number of rows registered in the MetaFrame. If left at None, it will use the number
            of rows in the input dataframe.
        progress_bar:
            Whether to display a progress bar.
        config:
            A path or MetaConfig object that contains information about the variable specifications
            , defaults, etc. Variable specs in the config parameter will be overwritten by the
            var_specs parameter.

        Returns
        -------
        MetaFrame:
            Initialized metasyn metaframe.
        """
        if isinstance(var_specs, (str, pathlib.Path, MetaConfig)) and config is None:
            warn(
                "Supplying the configuration through var_specs is deprecated and will be removed"
                f" in metasyn version 2.0. Use config={var_specs} instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            config = var_specs
            var_specs = None
        # Parse the var_specs into a MetaConfig instance.
        if config is None:
            meta_config = MetaConfig([], plugins, defaults={"privacy": privacy})
        elif isinstance(config, (pathlib.Path, str)):
            meta_config = MetaConfig.from_toml(config)
        else:
            meta_config = config

        # var_specs overrules variable specifications in the configuration (file).
        if var_specs is not None:
            meta_config.update_varspecs(var_specs)

        if plugins is not None:
            meta_config.plugins = plugins  # type: ignore
        if privacy is not None:
            meta_config.defaults.privacy = privacy  # type: ignore

        if df is not None and not isinstance(df, pl.DataFrame):
            if isinstance(df, (str, pathlib.Path)):
                raise ValueError("Please provide a DataFrame as input, not a string or path.")
            df = pl.DataFrame(df)
        all_vars = []
        columns = df.columns if df is not None else []
        if df is not None:
            for col_name in (pbar := tqdm(columns, disable=not progress_bar, unit="variables")):
                desc = col_name[:5] + "…" + col_name[-6:] if len(col_name) > 11 else col_name
                pbar.set_description(f"{desc:>12}")
                var_spec = meta_config.get(col_name)
                var = MetaVar.fit(
                    df[col_name],
                    var_spec.dist_spec,
                    meta_config.plugins,
                    var_spec.privacy,
                    var_spec.prop_missing,
                    var_spec.description,
                )
                all_vars.append(var)

        # Data free columns to be appended
        for var_spec in meta_config.iter_var(exclude=columns):
            if not var_spec.data_free:
                raise ValueError(
                    f"Column with name '{var_spec.name}' not found and not declared as data_free."
                )
            distribution = meta_config.plugins.create(var_spec)
            prop_missing = 0.0 if var_spec.prop_missing is None else var_spec.prop_missing
            var = MetaVar(
                var_spec.name,
                var_spec.var_type,
                distribution,
                description=var_spec.description,
                prop_missing=prop_missing,
            )
            all_vars.append(var)
        if df is None:
            if meta_config.n_rows is None:
                raise ValueError(
                    "Please provide the number of rows in the configuration, or supply a DataFrame."
                )
            return cls(all_vars, meta_config.n_rows, file_format)
        n_rows = len(df) if n_rows is None else n_rows
        return cls(all_vars, n_rows, file_format)

    @classmethod
    def from_config(cls, meta_config: MetaConfig) -> MetaFrame:
        """Create a MetaFrame using a configuration, but without a DataFrame.

        Parameters
        ----------
        meta_config
            Configuration to be used for creating the new MetaFrame.

        Returns
        -------
            A created MetaFrame.
        """
        return cls.fit_dataframe(None, config=meta_config)

    def to_dict(self) -> Dict[str, Any]:
        """Create dictionary with the properties for recreation."""
        self_dict = {
            "gmf_version": "1.1",
            "n_rows": self.n_rows,
            "n_columns": self.n_columns,
            "provenance": {
                "created by": {
                    "name": "metasyn",
                    "version": version("metasyn"),
                },
                "creation time": datetime.now().isoformat(),
            },
            "file_format": self.file_format,
            "vars": [var.to_dict() for var in self.meta_vars],
        }
        if self.file_format is None:
            self_dict.pop("file_format")
        return self_dict

    def __getitem__(self, key: Union[int, str]) -> MetaVar:
        """Return meta var either by variable name or index."""
        if isinstance(key, int):
            # If the key is an integer, give the ith variable.
            return self.meta_vars[key]
        if isinstance(key, str):
            # If the key is a string, return the first variable with that name.
            for var in self.meta_vars:
                if var.name == key:
                    return var
            raise KeyError(f"Cannot find variable '{key}'")
        raise TypeError(f"Cannot get item for key '{key}'")

    def __str__(self) -> str:
        """Return an easy to read formatted string for the metaframe."""
        vars_formatted = "\n".join(
            f"Column {i + 1}: {str(var)}" for i, var in enumerate(self.meta_vars)
        )
        return f"# Rows: {self.n_rows}\n# Columns: {self.n_columns}\n\n{vars_formatted}\n"

    @property
    def file_format(self) -> Optional[dict[str, Any]]:
        return self._file_format

    @file_format.setter
    def file_format(self, new_file_format: Union[None, dict[str, Any], BaseFileInterface]):
        if isinstance(new_file_format, BaseFileInterface):
            out_file_format: Optional[dict[str, Any]] = new_file_format.to_dict()
        else:
            out_file_format = new_file_format

        self._file_format = out_file_format

    @property
    def descriptions(self) -> dict[str, str]:
        """Return the descriptions of the columns."""
        return {
            var.name: var.description
            for var in self.meta_vars
            if var.name is not None and var.description is not None
        }

    @descriptions.setter
    def descriptions(self, new_descriptions: Union[dict[str, str], Sequence[str]]):
        if isinstance(new_descriptions, dict):
            for var_name, new_desc in new_descriptions.items():
                self[var_name].description = new_desc
        else:
            assert len(new_descriptions) == self.n_columns, (
                "Descriptions need to be either a dict or a "
                "sequence with the length of the number of variables."
            )
            for i_desc, new_desc in enumerate(new_descriptions):
                self[i_desc].description = new_desc

    def save(self, fp: Optional[Union[pathlib.Path, str]], validate: bool = True) -> None:
        """Serialize and save the MetaFrame to a JSON or TOML file, following the GMF format.

        Optionally, validate the saved JSON file against the JSON schema(s) included in the
        package. A TOML cannot be validated against a schema currently.

        Parameters
        ----------
        fp:
            File to write the metaframe to.
        validate:
            Validate the JSON file with a schema. If the file is a TOML file, then this will
            be ignored.
        """
        if fp is None:
            self.save_json(fp, validate)
            return
        fp_path = Path(fp)
        if fp_path.suffix == ".toml":
            self.save_toml(fp, validate)
        else:
            self.save_json(fp, validate)

    @classmethod
    def load(cls, fp: Union[pathlib.Path, str], validate: bool = True) -> MetaFrame:
        """Read a MetaFrame from a JSON or TOML GMF file.

        Optionally, validate the saved JSON file against the JSON schema(s) included in the
        package. A TOML cannot be validated against a schema currently.

        Parameters
        ----------
        fp:
            Path to read the data from.
        validate:
            Validate the JSON file with a schema. If the file is a TOML file, then this will
            be ignored.

        Returns
        -------
        MetaFrame:
            A restored MetaFrame from the file.
        """
        fp_path = Path(fp)
        if fp_path.suffix == ".toml":
            return cls.load_toml(fp, validate)
        else:
            return cls.load_json(fp, validate)

    def save_json(self, fp: Optional[Union[pathlib.Path, str]], validate: bool = True) -> None:
        """Serialize and save the MetaFrame to a JSON file, following the GMF format.

        Optionally, validate the saved JSON file against the JSON schema(s) included in the
        package.

        Parameters
        ----------
        fp:
            File to write the metaframe to.
        validate:
            Validate the JSON file with a schema.
        """
        self_dict = _jsonify(self.to_dict())
        if validate:
            validate_gmf_dict(self_dict)
        if fp is None:
            print(json.dumps(self_dict, indent=4))
        else:
            with open(fp, "w", encoding="utf-8") as f:
                json.dump(self_dict, f, indent=4)

    @classmethod
    def load_json(cls, fp: Union[pathlib.Path, str], validate: bool = True) -> MetaFrame:
        """Read a MetaFrame from a JSON file.

        Parameters
        ----------
        fp:
            Path to read the data from.
        validate:
            Validate the JSON file with a schema.

        Returns
        -------
        MetaFrame:
            A restored MetaFrame from the file.
        """
        with open(fp, "r", encoding="utf-8") as f:
            self_dict = json.load(f)

        if validate:
            validate_gmf_dict(self_dict)

        n_rows = self_dict["n_rows"]
        meta_vars = [MetaVar.from_dict(d) for d in self_dict["vars"]]
        return cls(meta_vars, n_rows, self_dict.get("file_format"))

    def to_json(self, fp: Union[pathlib.Path, str], validate: bool = True) -> None:
        """Export, deprecated method, use Metaframe.save_json instead."""
        warn(
            "to_json method of MetaFrame is deprecated and will be removed in the future, "
            "Use MetaFrame.save_json or MetaFrame.save instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.save_json(fp, validate)

    def export(self, fp: Union[pathlib.Path, str], validate: bool = True) -> None:
        """Export, deprecated method, use Metaframe.save instead."""
        warn(
            "Export method of MetaFrame is deprecated and will be removed in the future, "
            "Use MetaFrame.save_json or MetaFrame.save instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.save_json(fp, validate)

    @classmethod
    def from_json(cls, fp: Union[pathlib.Path, str], validate: bool = True) -> MetaFrame:
        """Import, deprecated method, use Metaframe.load_json instead."""
        warn(
            "MetaFrame.from_json is deprecated and will be removed in the future, "
            "use MetaFrame.load_json or MetaFrame.load instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.load_json(fp, validate)

    @no_type_check
    def save_toml(self, fp: Optional[Union[pathlib.Path, str]], validate: bool = True) -> None:
        try:
            import tomlkit  # noqa: PLC0415
        except ImportError:
            raise ValueError(
                "Please install tomlkit (pip install tomlkit) to enable support for saving to toml."
            )
        self_dict = _jsonify(self.to_dict())
        if validate:
            validate_gmf_dict(self_dict)

        doc = tomlkit.loads(tomlkit.dumps(self_dict))
        doc["n_rows"].comment("Number of rows")
        doc["n_columns"].comment("""Number of columns

# This is a metadata file with (limited) statistical information about each column separately in
# a dataset. No information about correlations or other relationships between columns is included.
# This file can be used to generate privacy-conscious synthetic data, which consequently has zero
# expected correlations and relationships between columns.
# For each column, the statistics can be either manually specified or estimated from real data.
# This information, including how the estimation was done, is shown in the metadata below.
#
# For more information, see https://github.com/sodascience/metasyn
""")
        for i in range(self.n_columns):
            var = self.meta_vars[i]
            doc["vars"][i].comment(f"Metadata for column with name {var.name}")
            fmi = round(self.n_rows * (1 - var.prop_missing))
            doc["vars"][i]["prop_missing"].comment(
                f"Fraction of missing values, remaining: {fmi} values"
            )
            # The below comment does not work, a tomlkit bug?
            # doc["vars"][i]["distribution"]["unique"].add(tomlkit.comment(
            # "Whether to generate unique values or not"))
            parameter_comments = []
            multi_default = (
                var.distribution.matches_name("multinoulli")
                and len(var.distribution.labels)
                == len(var.distribution.default_distribution(var_type=var.var_type).labels)
                and np.all(
                    var.distribution.labels == var.distribution.default_distribution(
                        var_type=var.var_type).labels
                )
            )
            if "parameters" in var.creation_method:
                parameters = ", ".join(var.creation_method["parameters"])
                parameter_comments.append(
                    f"The parameters {parameters} for column '{var.name}' were "
                    "manually set by the user, no data was (directly) used."
                )
            elif var.distribution.matches_name("multinoulli") and multi_default:
                parameter_comments.append(
                    "This mulinoulli distribution is the default one, no data was used."
                )
            elif "privacy" in var.creation_method:
                privacy = get_privacy(**var.creation_method["privacy"])
                parameter_comments.append(privacy.comment(var))
            if var.distribution.matches_name("multinoulli") and not multi_default:
                counts = (var.distribution.probs * (1 - var.prop_missing) * self.n_rows).round()
                parameter_comments.append(f"Counts: {counts.astype(int)}\n")
            par_comment = "\n# ".join(parameter_comments) + "\n\n"
            doc["vars"][i]["distribution"]["parameters"].add(tomlkit.comment(par_comment))
        if fp is None:
            print(tomlkit.dumps(doc))
        else:
            with open(fp, "w", encoding="utf-8") as f:
                tomlkit.dump(doc, f)

    @classmethod
    def load_toml(cls, fp: Union[pathlib.Path, str], validate: bool = True) -> MetaFrame:
        with open(fp, "rb") as f:
            self_dict = tomllib.load(f)

        if validate:
            validate_gmf_dict(self_dict)

        n_rows = self_dict["n_rows"]
        meta_vars = [MetaVar.from_dict(d) for d in self_dict["vars"]]
        return cls(meta_vars, n_rows, self_dict.get("file_format"))

    def synthesize(
        self,
        n: Optional[int] = None,
        seed: Optional[int] = None,
        progress_bar: bool = True,
    ) -> pl.DataFrame:
        """Create a synthetic Polars dataframe.

        Parameters
        ----------
        n:
            Number of rows to generate, if None, use number of rows in original dataframe.
        seed:
            Seed value for the internal random number generator. Set this to ensure reproducibility.
        progress_bar:
            Whether to display a progress bar.

        Returns
        -------
        polars.DataFrame:
            Dataframe with the synthetic data.
        """
        if n is None:
            if self.n_rows is None:
                raise ValueError(
                    "Cannot synthesize DataFrame, since number of rows is unknown."
                    "Please specify the number of rows to synthesize."
                )
            n = self.n_rows

        if seed is not None:
            set_global_seeds(seed)

        synth_dict = {}
        for var in (pbar := tqdm(self.meta_vars, disable=not progress_bar, unit="variables")):
            desc = var.name[:5] + "…" + var.name[-6:] if len(var.name) > 11 else var.name
            pbar.set_description(f"{desc:>12}")
            synth_dict[var.name] = var.draw_series(n, seed=None, progress_bar=progress_bar)

        return pl.DataFrame(synth_dict)

    def write_synthetic(
        self,
        file_name: Union[None, Path, str] = None,
        n: Optional[int] = None,
        seed: Optional[int] = None,
        file_format: Union[None, dict, BaseFileInterface] = None,
        overwrite: bool = False,
    ):
        """Write a synthetic dataset to a file.

        To write a synthetic dataset, by default it will try to create a file that has
        the same format as the original one. For example, if the separator of the CSV
        file was a comma, then it will write the synthetic data with the same separator.
        If the file format is not available (GMF files with older versions of metasyn
        or custom file interfaces), then you will have to supply your own file interface.

        Parameters
        ----------
        file_name:
            The filename to write the synthetic data to, by default None in which case
            the same filename will be used as for the original filename if available.
        n:
            Number of rows to be written for the new synthetic file, by default None
            in which case the number of rows of the original dataset will be used.
        seed:
            Set the seed for creating the synthetic dataset, by default None
        file_format:
            File format that determines how the file will be written. This is a dictionary
            that can be created by a file interface with the
            :meth:`metasyn.fileinterface.BaseFileInterface.to_dict` method. Example file interface
            classes are :class:`metasyn.fileinterface.CsvFileReader` and
            class:`metasyn.fileinterface.SavFileReader`. By default the file_format is None,
            in which case the file interface from the GMF file will be used, otherwise an error
            will be thrown.

        Raises
        ------
        ValueError:
            If the file format is None, and the MetaFrame object itself does not have a file format
            either.

        """
        if file_format is not None:
            if isinstance(file_format, BaseFileInterface):
                file_format = file_format.to_dict()
            if self.file_format is not None:
                if self.file_format["file_interface_name"] != file_format["file_interface_name"]:
                    warn(
                        "Writing the synthetic file with a different format as the original "
                        f"dataset. Original: {self.file_format['file_interface_name']}, "
                        f"Synthetic: {file_format['file_interface_name']}"
                    )
            self.file_format = file_format  # type: ignore
        if self.file_format is None:
            raise ValueError(
                "Cannot write synthetic dataset without file handler."
                " Use write_synthetic(..., file_format=your_file_handler.to_dict())"
            )
        file_handler = file_interface_from_dict(self.file_format)
        file_handler.check_filename(file_name, overwrite=overwrite)  # Check filename before synth
        syn_df = self.synthesize(n, seed)
        file_handler.write_file(syn_df, file_name, overwrite=overwrite)

    def __repr__(self) -> str:
        """Return the MetaFrame as it would be output to JSON."""
        if len(self.meta_vars) <= 3:
            var_str = ", ".join(var.__repr__() for var in self.meta_vars)
        else:
            var_str = ", ".join(var.__repr__() for var in self.meta_vars[:2])
            var_str += ", ..."
        return f"MetaFrame: size = ({self.n_rows} x {self.n_columns}) <{var_str}>"


def _jsonify(data):
    if isinstance(data, (list, tuple)):
        return [_jsonify(d) for d in data]
    if isinstance(data, dict):
        return {key: _jsonify(value) for key, value in data.items()}

    if isinstance(data, (np.int8, np.int16, np.int32, np.int64)):
        return int(data)
    if isinstance(data, np.float32):
        return float(data)
    if isinstance(data, np.ndarray):
        return _jsonify(data.tolist())
    return data



"""Module serving as the basis for all metasyn distributions.

The base module contains the ``BaseDistribution`` class,
which is the base class for all distributions.
It also contains the ``ScipyDistribution`` class,
which is a specialized base class for distributions that are built on top of
SciPy's statistical distributions.

Additionally it contains the ``UniqueDistributionMixin`` class,
which is a mixin class that can be used to make a distribution unique
(i.e., one that does not contain duplicate values).

Finally it contains the ``metadist()`` decorator,
which is used to set the class attributes of a distribution.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from importlib import metadata
from typing import Any, Optional, Sequence, Union

import numpy as np
import polars as pl
from numpy import typing as npt

from metasyn.privacy import BasePrivacy
from metasyn.util import get_var_type


def convert_to_series(values: Union[npt.NDArray, pl.Series]) -> pl.Series:
    """Convert list or pandas series to polars series."""
    if not isinstance(values, (pl.Series, np.ndarray)):
        values = pl.Series(values)
    if isinstance(values, pl.Series):
        series = values.drop_nulls()
    else:
        series_array = np.array(values)
        series_array = series_array[~np.isnan(series_array)]
        series = pl.Series(series_array)
    return series

class BaseFitter(ABC):
    """Base class for fitters."""

    distribution: type[BaseDistribution]
    privacy_type: str = "none"
    version: str = "1.0"
    var_type: Union[str, list[str]] = "unknown"
    plugin: str = "unknown"
    plugin_version: str = "1.0"

    def __init__(self, privacy: BasePrivacy):
        self.privacy = privacy

    def fit(self, values: Union[npt.NDArray, pl.Series]) -> BaseDistribution:
        pl_series = convert_to_series(values)
        if len(pl_series) == 0:
            return self.distribution.default_distribution(get_var_type(pl_series))
        return self._fit(pl_series)

    @abstractmethod
    def _fit(self, series: pl.Series) -> BaseDistribution:
        raise NotImplementedError

    @classmethod
    def matches_name(cls, name: str) -> bool:
        """Check whether the name matches the fitter.

        Parameters
        ----------
        name: str
            Name to match to the fitter.

        Returns
        -------
        bool:
            Whether the name matches.
        """
        assert cls.distribution.name != "unknown", f"Internal error in class {cls.__name__}"
        return name in (cls.distribution.name.split(".")[1],
                        cls.distribution.name,
                        cls.__name__,
                        )

    @classmethod
    def provides_var_type(cls, var_type: str) -> bool:
        if isinstance(cls.var_type, str):
            return cls.var_type == var_type
        return var_type in cls.var_type

    def to_dict(self):
        return {
            "name": self.__class__.__name__,
            "privacy": self.privacy.to_dict(),
            "version": self.version,
            "plugin": self.plugin,
            "plugin_version": self.plugin_version,
        }

class BaseDistribution(ABC):
    """Abstract base class to define a distribution.

    All distributions should be derived from this class, and should implement
    the following methods:
    :meth:`~_fit`,
    :meth:`~draw`,
    :meth:`~_param_dict`,
    :meth:`~_param_schema`,
    :meth:`~default_distribution`
    and ``__init__``.
    """

    name: str = "unknown"
    """The identifier for the implemented distribution"""
    var_type: Union[str, Sequence[str]] = "unknown"
    """The variable type of the distribution"""
    unique: bool = False
    """Whether the distribution creates only unique values"""
    version: str = "1.0"
    """Version of the implemented distribution"""

    @abstractmethod
    def draw(self) -> object:
        """Draw a random element from the fitted distribution."""

    def draw_reset(self) -> None:
        """Reset the drawing of elements to start again."""

    @property
    def _params_formatted(self) -> str:
        return "\n".join(
            f"\t- {param}: {value}" for param,
            value in self._param_dict().items()
        )

    def __str__(self) -> str:
        """Return an easy to read formatted string for the distribution."""
        return (
            f"- Type: {self.name}\n"
            f"- Parameters:\n"
            f"{self._params_formatted}\n"
        )

    @abstractmethod
    def _param_dict(self):
        """Get dictionary with the parameters of the distribution."""

    def to_dict(self) -> dict:
        """Convert the distribution to a dictionary."""
        return {
            "name": self.name,
            "version": self.version,
            "class_name": self.__class__.__name__,
            "unique": self.unique,
            "parameters": deepcopy(self._param_dict()),
        }

    @classmethod
    @abstractmethod
    def _param_schema(cls):
        """Get schema for the parameters of the distribution."""

    @classmethod
    def schema(cls) -> dict:
        """Create sub-schema to validate GMF file."""
        return {
            "type": "object",
            "properties": {
                "name": {"const": cls.name},
                "version": {"type": "string"},
                "class_name": {"const": cls.__name__},
                "unique": {"const": cls.unique},
                "parameters": {
                    "type": "object",
                    "properties": cls._param_schema(),
                    "required": list(cls.default_distribution()._param_dict())
                }
            },
            "required": ["name", "class_name", "parameters"]
        }

    @classmethod
    def from_dict(cls, dist_dict: dict) -> BaseDistribution:
        """Create a distribution from a dictionary."""
        return cls(**dist_dict["parameters"])

    def information_criterion(self, values: Union[pl.Series, npt.NDArray]) -> float: # noqa: ARG002
        """Get the BIC value for a particular set of values.

        Parameters
        ----------
        values: array_like
            Values to determine the BIC value of.
        """
        return 0.0

    @classmethod
    def matches_name(cls, name: str) -> bool:
        """Check whether the name matches the distribution.

        Parameters
        ----------
        name: str
            Name to match to the distribution.

        Returns
        -------
        bool:
            Whether the name matches.
        """
        assert cls.name != "unknown", f"Internal error in class {cls.__name__}"
        return name in (cls.name.split(".")[1],
                        cls.name,
                        cls.__name__,
                        )

    @classmethod
    def provides_var_type(cls, var_type: str) -> bool:
        if isinstance(cls.var_type, str):
            return cls.var_type == var_type
        return var_type in cls.var_type

    @classmethod
    @abstractmethod
    def default_distribution(cls, var_type: Optional[str] = None) -> BaseDistribution:
        """Get a distribution with default parameters."""
        return cls()

    def draw_list(self, n: int) -> list:
        """Draw a list of values from the distribution.

        Parameters
        ----------
        n:
            Number of items to draw from the distribution.

        Raises
        ------
        NotImplementedError:
            If the distribution hasn't implemented a draw_list.

        Returns
        -------
            List of values.
        """
        raise NotImplementedError()

def metadist(
        name: Optional[str] = None,
        var_type: Optional[Union[str, list[str]]] = None,
        unique: Optional[bool] = None,
        version: Optional[str] = None):
    """Decorate class to create a distribution with the right properties.

    Parameters
    ----------
    name:
        Name that identifies the distribution uniquely, e.g. core.uniform, core.regex.
        The name should use a period (.) so that the first part is the namespace (e.g. core),
        and the second part the name of the distribution.
    var_type:
        Variable type of the distribution, e.g. continuous, categorical, string.
    unique:
        Whether the distribution is unique or not.
    version:
        Version of the distribution. Increment this to ensure that compatibility is
        properly handled.

    Returns
    -------
    cls:
        Class with the appropriate class variables.
    """
    def _wrap(cls):
        if name is not None:
            cls.name = name
        if var_type is not None:
            cls.var_type = var_type
        if unique is not None:
            cls.unique = unique
        if version is not None:
            cls.version = version
        if cls.__doc__ is None:
            return cls
        cls.__doc__ = cls.__doc__.rstrip(" ")
        if not cls.__doc__.endswith("\n"):
            cls.__doc__ += "\n"
        cls.__doc__ += f"""
    Attributes
    ----------
    name:
        {cls.name}
    unique:
        {cls.unique}
    version:
        {cls.version}
    var_type:
        {cls.var_type}
    """
        return cls

    return _wrap


def metafit(
        distribution: Optional[type[BaseDistribution]] = None,
        var_type: Optional[Union[str, list[str]]] = None,
        version: Optional[str] = None,
        privacy_type: Optional[str] = None,
        plugin: Optional[str] = None,
        plugin_version: Optional[str] = None):
    """Decorate class to create a fitter with the correct class attributes.

    Parameters
    ----------
    distribution:
        Class that the fitter will return after a succesful fit.
    var_type:
        Variable type(s) that the fitter implements, e.g. continuous, categorical, string.
    version:
        Version of the fitter. Increment this to ensure that compatibility is
        properly handled.
    privacy_type:
        Privacy class/implementation of the fitter.
    plugin:
        Name of the plugin for the fitter or builtin (if part of metasyn itself).
    plugin_version:
        Version of the plugin used.

    Returns
    -------
    cls:
        Class with the appropriate class variables.
    """
    def _wrap(cls):
        if distribution is not None:
            cls.distribution = distribution
        if var_type is not None:
            cls.var_type = var_type
        # if unique is not None:
            # cls.unique = unique
        if privacy_type is not None:
            cls.privacy_type = privacy_type
        if version is not None:
            cls.version = version
        if plugin is not None:
            cls.plugin = plugin
        if plugin_version is not None:
            cls.plugin_version = plugin_version

        if cls.__doc__ is None:
            return cls
        cls.__doc__ = cls.__doc__.rstrip(" ")
        if not cls.__doc__.endswith("\n"):
            cls.__doc__ += "\n"
        cls.__doc__ += f"""
    Attributes
    ----------
    dist_class:
        {cls.distribution}
    version:
        {cls.version}
    var_type:
        {cls.var_type}
    privacy:
        {cls.privacy_type}
    """
        return cls

    return _wrap

def builtin_fitter(
        distribution: Optional[type[BaseDistribution]] = None,
        var_type: Optional[Union[str, list[str]]] = None,
        version: Optional[str] = None,
        privacy_type: Optional[str] = None):
    """Decorate builtin fitters.

    Parameters
    ----------
    distribution:
        Class that the fitter will return after a succesful fit.
    var_type:
        Variable type(s) that the fitter implements, e.g. continuous, categorical, string.
    version:
        Version of the fitter. Increment this to ensure that compatibility is
        properly handled.
    privacy_type:
        Privacy class/implementation of the fitter.

    Returns
    -------
    cls:
        Class with the appropriate class variables.
    """
    __version__ = metadata.version("metasyn")
    def _wrap(cls):
        wrapper = metafit(distribution, var_type, version, privacy_type,
                          plugin="builtin", plugin_version=__version__)
        return wrapper(cls)

    return _wrap

class ScipyDistribution(BaseDistribution):
    """Base class for numerical distributions using Scipy.

    This base class makes it easy to implement new numerical
    distributions. It can also be used for non-Scipy distributions,
    provided the distribution implements `logpdf`, `rvs` and `fit` methods.
    """

    @property
    def n_par(self) -> int:
        """int: Number of parameters for distribution."""
        return len(self.par)

    def __getattr__(self, attr: str):
        """Get attribute for easy access to parameters.

        Parameters
        ----------
        attr:
            Attribute to retrieve. If the attribute is a parameter
            name, then retrieve that parameter, otherwise use the default
            implementation for getting the attribute.

        Returns
        -------
        object:
            Parameter or attribute.
        """
        if attr != "par" and attr in self.par:
            return self.par[attr]
        return object.__getattribute__(self, attr)

    def _param_dict(self):
        return self.par

    def draw(self):
        val = self.dist.rvs()
        if self.var_type == "discrete":
            return int(val)
        return val

    def draw_list(self, n: int) -> list:
        values = self.dist.rvs(n)
        if self.var_type == "discrete":
            values = values.astype(np.int64)
        return list(values)

    def information_criterion(self, values):
        vals = convert_to_series(values)
        if len(vals) == 0:
            return 2 * self.n_par
        return self._information_criterion(vals)

    def _information_criterion(self, values):
        return np.log(len(values)) * self.n_par - 2 * np.sum(self.dist.logpdf(values))


class ScipyFitter(BaseFitter):
    """Base fitter for scipy distributions."""

    def _fit(self, series):
        param = self.distribution.scipy_class.fit(series)  # type: ignore  # All derived classes have dist_class
        return self.distribution(*param)



@metadist(unique=True)
class UniqueDistributionMixin(BaseDistribution):
    """Mixin class to make unique version of base distributions.

    This mixin class can be used to extend base distribution classes, adding
    functionality that ensures generated values are unique. It overrides
    the `draw` method of the base class, adding a check to prevent duplicate
    values from being drawn. If a duplicate value is drawn, it retries up to
    1e5 times before raising a ValueError.

    The `UniqueDistributionMixin` is used in various unique metasyn distribution
    variations, such as `UniqueFakerDistribution` and `UniqueRegexDistribution`.
    """

    def __new__(cls, *args, **kwargs):  # noqa
        instance = super().__new__(cls)
        instance.key_set: set[Any] = set()
        return instance

    def draw_reset(self):
        self.key_set = set()

    def draw(self) -> object:
        n_retry = 0
        while n_retry < 1e5:
            new_val = super().draw()  # type: ignore
            if new_val not in self.key_set:
                self.key_set.add(new_val)
                return new_val
            n_retry += 1
        raise ValueError(f"Failed to draw unique string after {n_retry} tries.")

    def information_criterion(self, values): # noqa: ARG002
        return 9999999





"""Testing module for continuous distributions."""
import numpy as np
from pytest import mark
from scipy import stats

from metasyn.distribution.exponential import ExponentialFitter
from metasyn.distribution.normal import (
    ContinuousNormalFitter,
    ContinuousTruncatedNormalFitter,
    LogNormalFitter,
)
from metasyn.distribution.uniform import ContinuousUniformFitter
from metasyn.privacy import BasicPrivacy


@mark.parametrize(
    "lower,upper",
    [
        (0, 1),
        (-10, 2),
        (0.01, 0.02),
    ]
)
def test_uniform(lower, upper):
    """Test the continuous uniform distribution."""
    scale = upper-lower
    values = stats.uniform(loc=lower, scale=scale).rvs(100)
    dist = ContinuousUniformFitter(BasicPrivacy()).fit(values)
    assert dist.lower <= values.min()
    assert dist.upper >= values.max()
    assert dist.information_criterion(values) < 2*np.log(len(values))-200*np.log((upper-lower)**-1)
    assert isinstance(dist.draw(), float)


def test_uniform_bic():
    """Test whether out of bounds BIC works properly."""
    lower = 0
    upper = 1
    scale = upper-lower
    values = stats.uniform(loc=lower, scale=scale).rvs(1000)
    dist = ContinuousUniformFitter.distribution(0.5, 1.0)
    bic = dist.information_criterion(values)
    assert bic > 1000

@mark.parametrize(
    "mean,sd",
    [
        (0, 1),
        (-10, 2),
        (0.01, 0.02),
    ]
)
def test_normal(mean, sd):
    """Test the normal/Gaussian distribution."""
    values = stats.norm(loc=mean, scale=sd).rvs(1000)
    dist = ContinuousNormalFitter(BasicPrivacy()).fit(values)
    dist_uniform = ContinuousUniformFitter(BasicPrivacy()).fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert (dist.mean - mean)/sd < 0.5
    assert (dist.sd - sd)/sd < 0.5
    assert isinstance(dist.draw(), float)


@mark.parametrize(
    "mean,sd",
    [
        (0, 1),
        (-10, 2),
        (0.01, 0.02),
    ]
)
def test_log_normal(mean, sd):
    """Test the log normal distribution."""
    values = stats.lognorm(s=sd, loc=0, scale=np.exp(mean)).rvs(1000)
    dist = LogNormalFitter(BasicPrivacy()).fit(values)
    dist_uniform = ContinuousUniformFitter(BasicPrivacy()).fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert (dist.mean-mean)/sd < 1
    assert (dist.sd-sd)/sd < 1
    assert isinstance(dist.draw(), float)


@mark.parametrize(
    "lower,upper,mean,sd",
    [
        (-0.5, 0.5, 0, 0.5),
        (-10, -8, 0, 5),
        (-5, 3, 2, 2),
        (-0.01, 0.01, 0.01, 0.01),
    ]
)
def test_trunc_normal(lower, upper, mean, sd):
    """Test the truncated normal distribution."""
    a, b = (lower-mean)/sd, (upper-mean)/sd
    values = stats.truncnorm(a=a, b=b, loc=mean, scale=sd).rvs(5000)
    dist = ContinuousTruncatedNormalFitter(BasicPrivacy()).fit(values)
    dist_uniform = ContinuousUniformFitter(BasicPrivacy()).fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert isinstance(dist.draw(), float)


@mark.parametrize(
    "rate",
    [0.1, 10, 100, 234.1234]
)
def test_exponential(rate):
    """Test the exponential distribution."""
    values = stats.expon(loc=0, scale=1/rate).rvs(5000)
    dist = ExponentialFitter(BasicPrivacy()).fit(values)
    dist_uniform = ContinuousUniformFitter(BasicPrivacy()).fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert isinstance(dist.draw(), float)
    assert (dist.rate - rate)/rate < 0.1

def test_exponential_negative():
    dist = ExponentialFitter(BasicPrivacy()).fit([-1, -2, -3])
    assert dist.rate == ExponentialFitter.distribution.default_distribution().rate




"""Module to test the registry and registry list classes."""

import pytest
from pytest import mark, raises

from metasyn.distribution.base import BaseDistribution
from metasyn.distribution.categorical import MultinoulliDistribution
from metasyn.distribution.faker import FakerDistribution
from metasyn.distribution.normal import ContinuousNormalDistribution, DiscreteNormalDistribution
from metasyn.distribution.regex import RegexDistribution, UniqueRegexDistribution
from metasyn.distribution.uniform import ContinuousUniformDistribution, DiscreteUniformDistribution
from metasyn.registry import DistributionRegistry
from metasyn.varspec import DistributionSpec, VarSpec


@mark.parametrize("input", ["builtin", "fake-name"])
def test_plist_init(input):
    """Test whether the initialization works correctly."""
    if input == "fake-name":
        with pytest.raises(ValueError):
            DistributionRegistry.parse(input)
    else:
        plist = DistributionRegistry.parse(input)
        assert issubclass(plist.find_distribution("core.regex", var_type="string"),
                          BaseDistribution)


# @metadist(version="1.0")
# class UniformTest1(ContinuousUniformDistribution):
#     """Version 1.0 for legacy tests."""


# @metadist(version="1.1")
# class UniformTest11(UniformDistribution):
#     """Version 1.1 for legacy tests."""


# @metadist(version="2.0")
# class UniformTest2(UniformDistribution):
#     """Version 2.0 for legacy tests."""


# class CheckProvider(BuiltinDistributionProvider):
#     """Provider for legacy tests."""

#     distributions = [UniformTest2]
#     legacy_distributions = [UniformTest1, UniformTest11]


# class LegacyOnly(BuiltinDistributionProvider):
#     """Provider with deprecated uniform distributions."""

#     distributions = [MultinoulliDistribution]
#     legacy_distributions = [UniformTest1, UniformTest11, UniformTest2]


# def test_legacy():
#     """Test whether the legacy system/distributions work."""
#     plist = DistributionRegistry(CheckProvider)
#     assert issubclass(plist.find_distribution("core.uniform", var_type="continuous"), UniformTest2)
#     with pytest.warns():  # version 1.1 is deprecated
#         assert issubclass(
#             plist.find_distribution(
#                 "core.uniform", var_type="continuous", version="1.1"), UniformTest11)
#     with pytest.warns():  # version 1.0 is deprecated
#         assert issubclass(
#             plist.find_distribution("core.uniform", var_type="continuous", version="1.0"),
#             UniformTest1)
#     with pytest.warns():  # version 1.2 does not exist, so 1.1 should be chosen.
#         assert issubclass(
#             plist.find_distribution("core.uniform", var_type="continuous", version="1.2"),
#             UniformTest11)
#     with pytest.raises(ValueError):  # No version 0.x available
#         plist.find_distribution("core.uniform", var_type="continuous", version="0.9")
#     with pytest.warns():  # No version 2.1 available, but 2.0 is.
#         assert issubclass(
#             plist.find_distribution("core.uniform", var_type="continuous", version="2.1"),
#             UniformTest2)

#     plist = DistributionRegistry(LegacyOnly)
#     with pytest.warns():  # core.uniform is deprecated.
#         assert issubclass(
#             plist.find_distribution("core.uniform", var_type="continuous"),
#             UniformTest2)


@mark.parametrize(
    "dist_str,var_type,is_unique,dist",
    [
        ("multinoulli", "string", False, MultinoulliDistribution),
        ("multinoulli", "discrete", False, MultinoulliDistribution),
        ("uniform", "continuous", False, ContinuousUniformDistribution),
        ("uniform", "discrete", False, DiscreteUniformDistribution),
        ("normal", "continuous", False, ContinuousNormalDistribution),
        ("normal", "discrete", False, DiscreteNormalDistribution),
        ("faker", "string", False, FakerDistribution),
        ("regex", "string", False, RegexDistribution),
        ("regex", "string", True, UniqueRegexDistribution),
    ]
)
def test_find_distribution(dist_str, var_type, is_unique, dist):
    """Test whether distributions can be found using the registries."""
    registry_list = DistributionRegistry.parse("builtin")
    dist_class = registry_list.find_distribution(dist_str, var_type=var_type, unique=is_unique)
    assert dist == dist_class
    if "faker" in dist_str:
        with raises(ValueError):
            registry_list.find_distribution("this is not a distribution", "string")
    new_class = registry_list.find_distribution(dist_class.__name__, var_type=var_type,
                                                unique=is_unique)
    assert new_class == dist_class

def test_create_distribution():
    dist_spec = DistributionSpec("uniform", False, parameters={"lower": 10, "upper": 20})
    var_spec = VarSpec("test", dist_spec, var_type="continuous")
    registry_list = DistributionRegistry.parse("builtin")

    assert isinstance(registry_list.create(var_spec), ContinuousUniformDistribution)

    # Error with missing parameters
    var_spec.dist_spec.parameters.pop("lower")
    with pytest.raises(ValueError):
        registry_list.create(var_spec)

    # Error with unknown parameters
    with pytest.raises(TypeError):
        var_spec.dist_spec.parameters["unknown"] = 1
        registry_list.create(var_spec)

    # Error when name is not given.
    with pytest.raises(ValueError):
        var_spec.dist_spec.name = None
        registry_list.create(var_spec)



"""Testing module for the distributions that are builtin to metasyn directly."""
from pytest import mark, raises

from metasyn.distribution.base import metafit
from metasyn.distribution.uniform import ContinuousUniformDistribution, ContinuousUniformFitter
from metasyn.privacy import BasicPrivacy
from metasyn.registry import DistributionRegistry
from metasyn.testutils import check_distribution_registry, check_fitter


def test_builtin_registry():
    """Test the builtin distribution registry with the metasyn test functionality."""
    check_distribution_registry("builtin")


@mark.parametrize(
    "fitter", DistributionRegistry.parse("builtin").fitters
)
def test_dist_validation(fitter):
    """Check all distributions for correctness."""
    print(type(fitter), fitter)
    check_fitter(fitter, privacy=BasicPrivacy(), provenance="builtin")


class BadDistribution(ContinuousUniformDistribution):
    """Distribution that has a broken schema."""

    def schema():
        return "[{"

@metafit(distribution=BadDistribution)
class BadFitter(ContinuousUniformFitter):
    pass

def test_schema_val():
    """Check whether a distribution with a broken schema will trigger the error."""
    with raises(ValueError):
        check_fitter(BadFitter, "none", "builtin")



"""Module to test the MetaVar object.

Most of the test functions try to infer the correct distribution from a series.
"""
import datetime as dt
import json
import string

import numpy as np
import pandas as pd
import polars as pl
import pytest
from pytest import mark, raises

from metasyn.distribution.categorical import MultinoulliDistribution
from metasyn.distribution.normal import (
    ContinuousNormalDistribution,
    ContinuousTruncatedNormalDistribution,
)
from metasyn.distribution.regex import RegexDistribution, UniqueRegexDistribution
from metasyn.distribution.uniform import ContinuousUniformDistribution, DiscreteUniformDistribution
from metasyn.distribution.uniquekey import UniqueKeyDistribution
from metasyn.metaframe import _jsonify
from metasyn.util import get_var_type
from metasyn.var import MetaVar


def _series_drop_nans(series):
    if isinstance(series, pl.Series):
        return series.drop_nulls()
    return series.dropna()


def _series_element_classname(series, all_nan):
    if not all_nan:
        series = _series_drop_nans(series)
    if isinstance(series, pl.Series):
        return series[0].__class__.__name__
    return series.iloc[0].__class__.__name__


def check_var(series, var_type, temp_path, all_nan=False):
    """Checker for all variables whether series are inferred correctly."""
    def check_similar(series_a, series_b):
        assert isinstance(series_a, (pd.Series, pl.Series))
        assert isinstance(series_b, (pd.Series, pl.Series))
        assert len(series_a) == len(series_b)
        base_type_a = _series_element_classname(series_a, all_nan)
        base_type_b = _series_element_classname(series_b, all_nan)
        if type(series_a) is type(series_b):
            assert base_type_a == base_type_b
        has_nans_a = len(series_a) - len(_series_drop_nans(series_a)) > 0
        has_nans_b = len(series_b) - len(_series_drop_nans(series_b)) > 0
        assert has_nans_a == has_nans_b

    def check_random_draw(var, n_series):
        series_1 = var.draw_series(n_series, 1234)
        series_2 = var.draw_series(n_series, 1234)
        assert all(_series_drop_nans(series_1) == _series_drop_nans(series_2))

    assert isinstance(series, (pd.Series, pl.Series))

    var = MetaVar.fit(series)
    new_series = var.draw_series(len(series), 5123)
    check_random_draw(var, len(series))
    check_similar(series, new_series)
    assert var.var_type == var_type
    assert var_type in var.distribution.var_type
    assert var.creation_method["created_by"] == "metasyn"


    new_var = MetaVar.from_dict(var.to_dict())
    with raises(ValueError):
        var_dict = var.to_dict()
        var_dict.update({"type": "unknown"})
        print(MetaVar.from_dict(var_dict))

    with raises(ValueError):
        var_dict = var.to_dict()
        var_dict["distribution"].update({"name": "unknown"})
        MetaVar.from_dict(var_dict)

    newer_series = new_var.draw_series(len(series), 6789)
    check_similar(newer_series, series)

    assert type(new_var) is type(var)
    assert new_var.dtype == var.dtype
    assert var_type == new_var.var_type

    # Write to JSON file and read it back
    tmp_fp = temp_path / "tmp.json"
    with open(tmp_fp, "w") as f:
        json.dump(_jsonify(var.to_dict()), f)

    with open(tmp_fp, "r") as f:
        new_var = MetaVar.from_dict(json.load(f))
    check_similar(series, new_var.draw_series(len(series), 8234))

    assert type(new_var) is type(var)
    assert new_var.dtype == var.dtype
    assert new_var.var_type == var_type
    assert new_var.creation_method["created_by"] == "metasyn"

    return new_series


@mark.parametrize(
    "series",
    [
        pd.Series(np.random.choice(["a", "b", "c", None], size=100), dtype="category"),
        pl.Series(np.random.choice(["a", "b", "c", None], size=100).tolist(), dtype=pl.Categorical)
    ]
)
def test_categorical(tmp_path, series):
    """Test the inference of categorical variables."""
    new_series = check_var(series, "categorical", tmp_path)
    assert set(_series_drop_nans(series)) == set(np.unique(_series_drop_nans(new_series)))

@mark.parametrize("dtype", ["int8", "int16", "int32", "int64", "int"])
def test_integer(dtype, tmp_path):
    """Test the inference of integer variables."""
    series = pd.Series([np.random.randint(0, 10) for _ in range(300)], dtype=dtype)
    new_series = check_var(series, "discrete", tmp_path)
    assert new_series.min() >= 0
    assert new_series.max() < 10


@mark.parametrize("dtype", ["Int8", "Int16", "Int32", "Int64"])
def test_nullable_integer(dtype, tmp_path):
    """Test for pandas nullable integer data types (no applicable to polars)."""
    series = pd.Series([np.random.randint(0, 10) if np.random.rand() > 0.5 else None
                        for _ in range(500)], dtype=dtype)
    new_series = check_var(series, "discrete", tmp_path)
    assert new_series.min() >= 0
    assert new_series.max() < 10

@mark.parametrize(
    "dtype", [pl.Int8, pl.Int16, pl.Int32, pl.Int32, pl.Int64,
              pl.UInt8, pl.UInt16, pl.UInt32, pl.UInt64]
)
def test_polars_discrete(dtype, tmp_path):
    """Test for all different variables of polars discrete types."""
    series = pl.Series("ser", [np.random.randint(0, 10) for _ in range(100)], dtype=dtype)
    new_series = check_var(series, "discrete", tmp_path)
    assert new_series.min() >= 0
    assert new_series.max() < 10

@mark.parametrize(
    "dtype", [pl.Float32, pl.Float64]
)
def test_polars_continuous(dtype, tmp_path):
    """Test for all polars continuous data types."""
    series = pl.Series("ser", np.random.rand(100), dtype=dtype)
    new_series = check_var(series, "continuous", tmp_path)
    assert new_series.min() > -0.5
    assert new_series.max() < 1.5


@mark.parametrize(
    "dtype", [pl.String, pl.Utf8]
)
def test_polars_string(dtype, tmp_path):
    """Test polars string types."""
    series = pl.Series("series", string.printable, dtype=dtype)
    check_var(series, "string", tmp_path)

@mark.parametrize(
    "dtype", [pl.Categorical]
)
def test_polars_categorical(dtype, tmp_path):
    """Test whether polars categorical data types are parsed correctly."""
    series = pl.Series("series", string.printable, dtype=dtype)
    check_var(series, "categorical", tmp_path)

@mark.parametrize(
    "series_type",
    [pl.Series, pd.Series],
)
def test_float(tmp_path, series_type):
    """Test for continuous variables, pandas or polars."""
    np.random.seed(3727442)
    series = series_type([np.random.rand() for _ in range(10000)])
    new_series = check_var(series, "continuous", tmp_path)
    assert new_series.min() > 0
    assert new_series.max() < 1

    series = pd.Series(np.random.randn(1000))
    check_var(series, "continuous", tmp_path)


@mark.parametrize(
    "series_type",
    [pl.Series, pd.Series],
)
def test_string(tmp_path, series_type):
    """Test polars/pandas string type fitting."""
    series = series_type(np.random.choice(["a", "b", "c", None], size=100).tolist())
    new_series = check_var(series, "string", tmp_path)
    unq_values = set(np.unique(_series_drop_nans(series)))
    new_unq_values = set(np.unique(_series_drop_nans(new_series)))
    assert unq_values == new_unq_values

@mark.parametrize("dtype,var_type", [
    (pl.Float32, "continuous"),
    (pl.Float64, "continuous"),
    (pl.Int32, "discrete"),
    (pl.Int64, "discrete"),
    (pl.Categorical, "categorical"),
    (pl.Utf8, "string")])
def test_na(tmp_path, dtype, var_type):
    """Test the various NA type distributions."""
    series = pl.Series("x", [None, None, None], dtype=dtype)
    print(series)
    new_series = check_var(series, var_type, tmp_path, all_nan=True)
    assert all(x is None for x in new_series)


@mark.parametrize(
    "series_type",
    [pl.Series, pd.Series]
)
def test_bool(tmp_path, series_type):
    """Test whether booleans are correctly handled as categorical variables."""
    series = series_type(np.random.choice([True, False], size=100))
    check_var(series, "categorical", tmp_path)
    var = MetaVar.fit(series)
    new_series = var.draw_series(10, 1234)
    assert new_series.dtype == pl.Boolean


@mark.parametrize(
    "prop_missing",
    [-1, -0.1, 1.2],
)
def test_invalid_prop(prop_missing):
    """Ensure that errors are thrown for prop_missing outside of [0, 1]."""
    MetaVar("test", "discrete", DiscreteUniformDistribution.default_distribution())
    with raises(ValueError):
        MetaVar("test", "discrete", DiscreteUniformDistribution.default_distribution(),
                prop_missing=prop_missing)

@mark.parametrize(
    "series,var_type",
    [
        (pl.Series([1, 2, 3]), "discrete"),
        (pd.Series([1, 2, 3]), "discrete"),
        (pl.Series([1.0, 2.0, 3.0]), "continuous"),
        (pl.Series(["1", "2", "3"]), "string"),
        (pl.Series(["1", "2", "3"], dtype=pl.Categorical), "categorical"),
        (pl.Series([dt.time.fromisoformat("10:38:12"), dt.time.fromisoformat("12:52:11")]),
            "time"),
        (pl.Series([dt.datetime.fromisoformat("2022-07-15T10:39:36"),
                    dt.datetime.fromisoformat("2022-08-15T10:39:36")]),
            "datetime"),
        (pl.Series([dt.date.fromisoformat("1903-07-15"), dt.date.fromisoformat("1940-07-16")]),
            "date"),
    ]
)
def test_get_var_type(series, var_type):
    """Test get_var_type method of MetaVar."""
    assert get_var_type(series) == var_type


def test_unsupported_type():
    series = pl.Series([MetaVar])
    with pytest.raises(TypeError):
        get_var_type(series)


@mark.parametrize(
    "series",
    [pd.Series([np.random.rand() for _ in range(5000)]),
     pl.Series([np.random.rand() for _ in range(5000)])]
)
def test_manual_fit(series):
    """Test adding dist_spec to MetaVar.fit call."""
    var = MetaVar.fit(series)
    assert isinstance(var.distribution, (ContinuousUniformDistribution, ContinuousTruncatedNormalDistribution))
    var = MetaVar.fit(series, dist_spec={"name": "normal"})
    assert isinstance(var.distribution, ContinuousNormalDistribution)
    var = MetaVar.fit(series, dist_spec=ContinuousUniformDistribution)
    assert isinstance(var.distribution, ContinuousUniformDistribution)
    var = MetaVar.fit(series, dist_spec=ContinuousNormalDistribution(0, 1))
    assert isinstance(var.distribution, ContinuousNormalDistribution)
    with raises(TypeError):
        var.fit(10)


@mark.parametrize(
    "series",
    [pd.Series([pd.NA for _ in range(10)]),
     pl.Series([None for _ in range(10)])]
)
def test_na_zero(series):
    """Test whether fitting a series with only NA works."""
    var = MetaVar.fit(series)
    assert var.var_type == "continuous"
    assert var.prop_missing == 1.0


@mark.parametrize(
    "series",
    [pd.Series(np.array([pd.NA if i != 0 else 1.0 for i in range(10)])),
     pl.Series([None if i != 0 else 1.0 for i in range(10)])]
)
def test_na_one(series):
    """Test whether fitting a series with exactly one non-NA values works."""
    var = MetaVar.fit(series)
    assert var.var_type == "continuous"
    assert abs(var.prop_missing-0.9) < 1e7


@mark.parametrize(
    "series",
    [pd.Series(np.array([np.nan if i < 2 else 0.123*i for i in range(10)])),
     pl.Series([None if i < 2 else 0.123*i for i in range(10)])]
)
def test_na_two(series):
    """Test whether fitting a series with exactly two non-NA values works."""
    var = MetaVar.fit(series)
    assert var.var_type == "continuous"
    assert abs(var.prop_missing-0.8) < 1e7


@mark.parametrize(
    "series",
    [pd.Series(np.random.randint(0, 1000, size=500)),
     pl.Series(np.random.randint(0, 1000, size=500))]
)
def test_manual_unique_integer(series):
    """Test whether discrete uniform/unique key are correctly inferred."""
    var = MetaVar.fit(series)
    assert isinstance(var.distribution, DiscreteUniformDistribution)
    var = MetaVar.fit(series, dist_spec = {"unique": True})
    assert isinstance(var.distribution, UniqueKeyDistribution)


@mark.parametrize(
    "series",
    [
        pd.Series(["x213", "2dh2", "4k2kk"]),
        pl.Series(["x213", "2dh2", "4k2kk"]),
    ]
)
def test_manual_unique_string(series):
    """Test which regexmodel is used dependent on the uniqueness."""
    var = MetaVar.fit(series)
    assert isinstance(var.distribution, RegexDistribution)
    var = MetaVar.fit(series, dist_spec={"unique": True})
    assert isinstance(var.distribution, UniqueRegexDistribution)


@mark.parametrize(
    "series",
    [
        pl.Series([2]*100 + [4]*100 + [12394]*100),
        pl.Series([-123]*12 + [123]*12),
        pl.Series(["2"]*100 + ["5"]*100 + ["123"]*100, dtype=pl.Categorical)
    ]
)
def test_int_multinoulli(series):
    """Test whether the multinoulli distribution is correctly inferred."""
    var = MetaVar.fit(series)
    assert isinstance(var.distribution, MultinoulliDistribution)


def test_error_multinoulli():
    """Test whether incorrect probabilities give a warning/error."""
    with raises(ValueError):
        MultinoulliDistribution(["1", "2"], [-0.1, 1.1])
    with pytest.warns():
        MultinoulliDistribution(["1", "2"], [0.1, 0.2])



from pathlib import Path

from metasyn import MetaFrame, demo_dataframe
from metasyn.config import VarSpec

# example dataframe from polars website
df = demo_dataframe("fruit")

# set A to unique and B to not unique
specs = [
    VarSpec("ID", unique=True),
    VarSpec("B", unique=False),
]

# create MetaFrame
mf = MetaFrame.fit_dataframe(df, var_specs=specs)

# write to json
gmf_path = Path("examples", "gmf_files", "example_gmf_simple.json")
mf.save(gmf_path)

# then, export json from secure environment

# outside secure environment, load json
mf_out = MetaFrame.load_json(gmf_path)

# create a fake dataset
df_syn = mf_out.synthesize(10, seed=1234)



"""Module with tests for the command line interface."""
import json
from pathlib import Path

import jsonschema
import polars as pl
import pytest
from pytest import fixture, mark

import metasyn as ms
from metasyn import MetaFrame
from metasyn.__main__ import main
from metasyn.file import _AVAILABLE_FILE_INTERFACES
from metasyn.validation import validate_gmf_dict

TMP_DIR_PATH = None


@fixture(scope="module")
def tmp_dir(tmp_path_factory) -> Path:
    """Directory with a configuration, dataset and GMF file."""
    global TMP_DIR_PATH
    if TMP_DIR_PATH is None:
        # Create a temporary input file
        TMP_DIR_PATH = tmp_path_factory.mktemp("data")
        json_path = TMP_DIR_PATH / "titanic.json"
        csv_fp = Path("tests", "data", "titanic.csv")
        csv_dt = {
            "PassengerId": pl.Int64,
            "Survived": pl.Categorical,
            "Pclass": pl.Categorical,
            "Name": str,
            "Sex": pl.Categorical,
            "SibSp": pl.Categorical,
            "Parch": pl.Categorical,
            "Ticket": str,
            "Cabin": str,
            "Embarked": pl.Categorical,
            "Age": float,
            "Fare": float
        }
        data_frame, file_format = ms.read_csv(csv_fp, schema_overrides=csv_dt, n_rows=100)
        meta_frame = MetaFrame.fit_dataframe(data_frame,
                                             var_specs=[{"name": "PassengerId",
                                                        "distribution": {"unique": True}}],
                                             file_format=file_format)
        meta_frame.save_json(json_path)
        config_fp = TMP_DIR_PATH / "config.ini"
        with open(config_fp, "w") as handle:
            handle.write("""
[[var]]
name = "PassengerId"
distribution = {unique = true}

[[var]]
name = "Fare"
prop_missing = 0.2
distribution = {name = "lognormal"}
""")
    return TMP_DIR_PATH


@mark.parametrize("ext", [x.extensions[0] for x in _AVAILABLE_FILE_INTERFACES.values()])
def test_cli(tmp_dir, ext):
    """A simple integration test for creating synthetic data from a GMF file."""
    # create out file path with correct extension
    out_file = tmp_dir / f"titanic{ext}"

    # create command to run in subprocess with arguments
    cmd = [
        "synthesize",                     # the subcommand
        "-n 25",                          # only generate 25 samples
        tmp_dir / "titanic.json",         # the input file
        "-o",
        out_file,                         # the output file
        "--seed",
        str(1234),
    ]

    # Run the cli with different extensions
    main(cmd)
    assert out_file.is_file()
    if ext == ".csv":
        df = pl.read_csv(out_file)
        assert len(df) == 25

    main(["synthesize", tmp_dir / "titanic.json", "--preview"])

    # Check if errors are raised when a csv file is supplied.
    with pytest.raises(SystemExit):
        main(["synthesize", out_file, "-o", out_file])


def test_main_cli():
    main(["--version"])
    main(["--help"])
    with pytest.raises(SystemExit):
        main(["command_does_not_exit"])

@mark.parametrize("config", [True, False])
def test_create_meta(tmp_dir, config):
    """CLI test on creating metadata from a csv."""
    out_file = tmp_dir / "test.json"
    cmd = [
        "create-meta",                      # the subcommand
        Path("tests", "data", "titanic.csv"),  # the input file
        "-o",
        out_file                            # the output file
    ]
    if config:
        cmd.extend(["--config", Path(tmp_dir) / 'config.ini'])
    main(cmd)
    assert out_file.is_file()
    meta_frame = MetaFrame.load_json(out_file)
    assert len(meta_frame.meta_vars) == 12

    with pytest.raises(SystemExit):
        main(["create-meta", "-o", "some_file"])


def test_schema_list():
    """Test whether all plugins/schemas are listed."""
    cmd = [
        "schema",
        "--list"
    ]
    main(cmd)


def test_schema_gen(tmp_dir):
    """Test whether the metadata schemas can be created."""
    titanic_json = tmp_dir / "titanic.json"
    schema_file = tmp_dir / "schema.json"
    cmd = [
        "schema",
        "builtin",
        "-o",
        schema_file
    ]
    main(cmd)
    with open(schema_file, "r") as handle:
        json_schema = json.load(handle)

    with open(titanic_json, "r") as handle:
        gmf_dict = json.load(handle)
    validate_gmf_dict(gmf_dict)
    jsonschema.validate(gmf_dict, json_schema)

    with pytest.raises(SystemExit):
        main(["schema", "non-existent-plugin"])


def test_datafree(tmp_dir):
    """Test generating synthetic data from only a configuration file and no real data."""
    gmf_fp = tmp_dir / "gmf_out.json"
    syn_fp = tmp_dir / "test_out.csv"
    cmd = [
        "create-meta",                      # the subcommand
        "--config", Path("tests", "data", "no_data_config.toml"),
        "--output", gmf_fp,              # the output file
    ]
    main(cmd)
    meta_frame = MetaFrame.load_json(gmf_fp)
    assert meta_frame.n_rows == 100
    assert len(meta_frame.meta_vars) == 3
    cmd2 = [
        "synthesize",
        gmf_fp,
        "-o",
        syn_fp
    ]
    main(cmd2)
    df = pl.read_csv(syn_fp)
    assert list(df.columns) == ["PassengerId", "Name", "Cabin"]
    assert len(df) == 100

def test_custom_file_interface(tmp_dir):
    config_fp = Path("examples", "config_files", "file_interface.toml")
    input_fp = Path("tests", "data", "actually_a_csv_file.sav")
    out_gmf = tmp_dir / "temp.json"
    cmd = [
        "create-meta",
        input_fp,
        "--config", config_fp,
        "--output", out_gmf,
    ]
    main(cmd)
    meta_frame = MetaFrame.load_json(out_gmf)
    assert len(meta_frame.meta_vars) == 12



"""Module for testing discrete distributions."""
from math import fabs

import numpy as np
import pandas as pd
import polars as pl
from pytest import mark
from scipy import stats

from metasyn.distribution.normal import (
    DiscreteNormalFitter,
    DiscreteTruncatedNormalFitter,
)
from metasyn.distribution.poisson import PoissonFitter
from metasyn.distribution.uniform import DiscreteUniformFitter
from metasyn.distribution.uniquekey import UniqueKeyDistribution, UniqueKeyFitter
from metasyn.privacy import BasicPrivacy


@mark.parametrize(
    "data",
    [
        [1, 2, 3, 4, 5],
        [-399, 12, 1, 492],
        [np.random.randint(0, 1000) for _ in range(1000)],
    ]
)
@mark.parametrize(
    "series_type", [pd.Series, pl.Series]
)
def test_uniform(data, series_type):
    """Test discrete uniform distribution."""
    series = series_type(data)
    dist = DiscreteUniformFitter(BasicPrivacy()).fit(series)
    assert dist.lower == series.min() and dist.upper == series.max()+1
    drawn_values = set([dist.draw() for _ in range(1000)])
    if dist.upper - dist.lower < 5:
        assert len(drawn_values) == len(series)
    drawn_values = np.array(list(drawn_values))
    assert np.isclose(dist.information_criterion(drawn_values),
                      np.log(len(drawn_values))*2+2*len(drawn_values)*(np.log(dist.upper-dist.lower)))


@mark.parametrize(
    "data,better_than_uniform,consecutive",
    [
        ([1, 2, 3, 4, 5], True, True),
        ([5, 4, 3, 2, 1], True, False),
        ([2, 4, 5, 7, 10, 6], True, False),
        ([-3, 1, -5, 3, -2, 0], True, False),
        ([-129384, 2198384, 293, 1293840], False, False),
        ([1, 1, 2, 2, 3, 3], False, False)
    ]
)
@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_integer_key(data, better_than_uniform, consecutive, series_type):
    """Test unique key distribution (included increasing id's)."""
    series = series_type(data)
    dist = UniqueKeyFitter(BasicPrivacy()).fit(series)
    unif_dist = DiscreteUniformFitter(BasicPrivacy()).fit(series)
    assert dist.lower == series.min()
    assert dist.consecutive == consecutive
    assert better_than_uniform == (dist.information_criterion(series)
                                   < unif_dist.information_criterion(series))
    assert isinstance(dist, UniqueKeyDistribution)

    drawn_values = np.array([dist.draw() for _ in range(100)])
    if consecutive:
        assert np.all(drawn_values == np.arange(dist.lower, dist.lower+100))


def test_unique_key_ic_zero():
    """Check behavior of BIC with zero values."""
    dist = UniqueKeyDistribution.default_distribution()
    assert dist.information_criterion([]) == 4

@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_poisson(series_type):
    """Test Poisson distribution."""
    series = series_type(stats.poisson(mu=10).rvs(1000))
    dist = PoissonFitter(BasicPrivacy()).fit(series)
    dist_unif = DiscreteUniformFitter(BasicPrivacy()).fit(series)
    assert fabs(dist.rate - 10) < 1
    assert dist.information_criterion(series) < dist_unif.information_criterion(series)

@mark.parametrize(
    "lower,upper,mean,sd",
    [
        (-1, 1, 0, 0.5),
        (-10, -8, 0, 5),
        (-5, 3, 2, 2),
    ]
)
def test_trunc_normal(lower, upper, mean, sd):
    """Test discrete version of truncated normal distribution."""
    a, b = (lower-mean)/sd, (upper-mean)/sd
    values = pl.Series(stats.truncnorm(a=a, b=b, loc=mean, scale=sd).rvs(5000)).cast(pl.Int64)
    dist = DiscreteTruncatedNormalFitter(BasicPrivacy()).fit(values)
    dist_uniform = DiscreteUniformFitter(BasicPrivacy()).fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert isinstance(dist.draw(), int)


@mark.parametrize(
    "mean,sd",
    [
        (0, 1),
        (-10, 2),
        (100, 200),
    ]
)
def test_normal(mean, sd):
    """Test for discrete version of normal/Gaussian distribution."""
    values = pl.Series(stats.norm(loc=mean, scale=sd).rvs(1000)).cast(pl.Int64)
    dist = DiscreteNormalFitter(BasicPrivacy()).fit(values)
    dist_uniform = DiscreteUniformFitter(BasicPrivacy()).fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert (dist.mean - mean)/sd < 0.5
    assert (dist.sd - sd)/sd < 0.5
    assert isinstance(dist.draw(), int)



"""Package providing different distributions used in metasyn.

The package consists of several distribution modules, it also includes
``base`` module which forms the basis of all distributions.

Each distribution class provides methods for fitting the distribution to a
a series of values, and for generating synthetic data based on the fitted
distribution. Each distribution class also provides a way to calculate the
information criterion, used for selecting the best distribution for
a given set of values.
"""

from metasyn.distribution.categorical import MultinoulliDistribution, MultinoulliFitter
from metasyn.distribution.constant import (
    ContinuousConstantDistribution,
    ContinuousConstantFitter,
    DateConstantDistribution,
    DateConstantFitter,
    DateTimeConstantDistribution,
    DateTimeConstantFitter,
    DiscreteConstantDistribution,
    DiscreteConstantFitter,
    StringConstantDistribution,
    StringConstantFitter,
    TimeConstantDistribution,
    TimeConstantFitter,
)
from metasyn.distribution.exponential import ExponentialDistribution, ExponentialFitter
from metasyn.distribution.faker import (
    FakerDistribution,
    FakerFitter,
    UniqueFakerDistribution,
    UniqueFakerFitter,
)
from metasyn.distribution.freetext import FreeTextDistribution, FreeTextFitter
from metasyn.distribution.na import NADistribution, NAFitter
from metasyn.distribution.normal import (
    ContinuousNormalDistribution,
    ContinuousNormalFitter,
    ContinuousTruncatedNormalDistribution,
    ContinuousTruncatedNormalFitter,
    DiscreteNormalDistribution,
    DiscreteNormalFitter,
    DiscreteTruncatedNormalDistribution,
    DiscreteTruncatedNormalFitter,
    LogNormalDistribution,
    LogNormalFitter,
)
from metasyn.distribution.poisson import PoissonDistribution, PoissonFitter
from metasyn.distribution.regex import (
    RegexDistribution,
    RegexFitter,
    UniqueRegexDistribution,
    UniqueRegexFitter,
)
from metasyn.distribution.uniform import (
    ContinuousUniformDistribution,
    ContinuousUniformFitter,
    DateTimeUniformDistribution,
    DateTimeUniformFitter,
    DateUniformDistribution,
    DateUniformFitter,
    DiscreteUniformDistribution,
    DiscreteUniformFitter,
    TimeUniformDistribution,
    TimeUniformFitter,
)
from metasyn.distribution.uniquekey import UniqueKeyDistribution, UniqueKeyFitter

builtin_fitters = [
    DiscreteUniformFitter, ContinuousUniformFitter, DateUniformFitter, TimeUniformFitter,
    DateTimeUniformFitter,
    RegexFitter, UniqueRegexFitter,
    ContinuousConstantFitter, DiscreteConstantFitter, DateConstantFitter,
    DateTimeConstantFitter, TimeConstantFitter, StringConstantFitter,
    ExponentialFitter,
    MultinoulliFitter,
    FakerFitter, UniqueFakerFitter,
    FreeTextFitter,
    PoissonFitter,
    ContinuousNormalFitter, LogNormalFitter, DiscreteTruncatedNormalFitter,
    ContinuousTruncatedNormalFitter, DiscreteNormalFitter,
    UniqueKeyFitter,
    NAFitter,
]

__all__ = [
    "MultinoulliDistribution",
    "ExponentialDistribution",
    "LogNormalDistribution",
    "ContinuousNormalDistribution",
    "ContinuousTruncatedNormalDistribution",
    "ContinuousUniformDistribution",
    "DateUniformDistribution",
    "DateTimeUniformDistribution",
    "TimeUniformDistribution",
    "DiscreteNormalDistribution",
    "DiscreteTruncatedNormalDistribution",
    "DiscreteUniformDistribution",
    "PoissonDistribution",
    "UniqueKeyDistribution",
    "FakerDistribution",
    "FreeTextDistribution",
    "UniqueFakerDistribution",
    "NADistribution",
    "RegexDistribution",
    "UniqueRegexDistribution",
    "ContinuousConstantDistribution",
    "DiscreteConstantDistribution",
    "StringConstantDistribution",
    "DateConstantDistribution",
    "TimeConstantDistribution",
    "DateTimeConstantDistribution",
]



"""Testing module for fitting dataframes for both pandas and polars."""
from pathlib import Path
from random import random

import pandas as pd
import polars as pl
import pytest
from pytest import mark

from metasyn.demo.dataset import _AVAILABLE_DATASETS, _get_demo_class, demo_dataframe, demo_file
from metasyn.metaframe import MetaFrame
from metasyn.privacy import BasicPrivacy
from metasyn.registry import DistributionRegistry
from metasyn.var import MetaVar

dtypes = {
    "PassengerId": "int",
    "Survived": "category",
    "Pclass": "category",
    "Name": "string",
    "Sex": "category",
    "SibSp": "category",
    "Parch": "category",
    "Ticket": "string",
    "Cabin": "string",
    "Embarked": "category",
    "Age": "float",
    "Fare": "float",
}


def _read_csv(fp, dataframe_lib):
    if dataframe_lib == "pandas":
        df = pd.read_csv(fp, dtype=dtypes)
        return df.iloc[:100]
    else:
        df = pl.read_csv(fp, schema_overrides={x: pl.Categorical for x, x_type in dtypes.items()
                                               if x_type == "category"})
        return df[:100]


@mark.parametrize("dataframe_lib", ["polars", "pandas"])
def test_dataset(tmp_path, dataframe_lib):
    """Integration test that loads the CSV, and creates a MetaFrame from that."""
    titanic_fp = Path("tests", "data", "titanic.csv")
    tmp_fp = tmp_path / "tmp.json"
    df = _read_csv(titanic_fp, dataframe_lib)
    dataset = MetaFrame.fit_dataframe(
        df,
        var_specs=[
                {"name": "Name", "prop_missing": 0.5},
                {"name": "Ticket", "description": "test_description"},
                {"name": "Fare", "distribution": {"name": "normal"}},
                {"name": "PassengerId", "distribution": {"unique": True}},
             ])

    def check_dataset(dataset):
        assert dataset.n_columns == 12
        assert dataset.n_rows == 100
        sub_types = {
            "int": "discrete",
            "string": "string",
            "category": "categorical",
            "float": "continuous",
        }
        for col, type_name in dtypes.items():
            meta_var_type = sub_types[type_name]
            assert dataset[col].var_type == meta_var_type
        with pytest.raises(KeyError):
            dataset["unknown"]
        with pytest.raises(IndexError):
            dataset[20]
        with pytest.raises(TypeError):
            dataset[MetaVar]

        assert "Rows" in str(dataset)
        assert len(dataset.synthesize()) == dataset.n_rows
        assert len(dataset.synthesize(5)) == 5
        with pytest.raises(ValueError):
            dataset.n_rows = None
            dataset.synthesize()
        dataset.n_rows = 100

    check_dataset(dataset)
    dataset.save_json(tmp_fp)
    dataset = MetaFrame.load_json(tmp_fp)
    check_dataset(dataset)

    # Check the description functionality of metasyn.
    dataset.descriptions = {"Embarked": "Some description", "Sex": "Other description"}
    assert dataset.descriptions["Embarked"] == "Some description"
    with pytest.raises(AssertionError):
        dataset.descriptions = "12345"
    dataset.descriptions = {"Embarked": "New description"}
    assert dataset.descriptions["Embarked"] == "New description"
    assert dataset.descriptions["Sex"] == "Other description"
    if isinstance(df, pl.DataFrame):
        col_names = df.columns
    else:
        col_names = list(df)
    dataset.descriptions = col_names
    for name in col_names:
        print(name, dataset.descriptions[name])
        assert dataset.descriptions[name] == name

    assert isinstance(repr(dataset), str)

    # Check whether non-columns raise an error
    with pytest.raises(ValueError):
        dataset = MetaFrame.fit_dataframe(df, var_specs=[{"name": "unicorn", "prop_missing": 0.5}])


def test_distributions(tmp_path):
    """Create all available distributions and save a metaframe with it."""
    tmp_fp = tmp_path / "tmp.json"

    registry = DistributionRegistry.parse("builtin")
    var_type_set = set()
    for fitter in registry.fitters:
        if isinstance(fitter.var_type, str):
            var_type_set.add(fitter.var_type)
        else:
            var_type_set.update(fitter.var_type)
    for var_type in var_type_set:
        for fitter in registry.filter_fitters(BasicPrivacy(), var_type):
            var = MetaVar(name="None", var_type=var_type,
                          distribution=fitter.distribution.default_distribution(),
                          prop_missing=random())
            dataset = MetaFrame([var], n_rows=10)
            dataset.save_json(tmp_fp)

@mark.parametrize(
    "dataset_name", list(_AVAILABLE_DATASETS)
)
def test_demo_datasets(tmp_path, dataset_name):
    """Test all built-in demo datasets and see if they can be synthesized."""
    demo_fp = demo_file(dataset_name)
    demo_df = demo_dataframe(dataset_name)
    demo_class = _get_demo_class(dataset_name)

    assert demo_fp.is_file()
    assert isinstance(demo_df, pl.DataFrame)

    mf = MetaFrame.fit_dataframe(demo_df, var_specs=demo_class.var_specs)
    assert isinstance(mf, MetaFrame)

    tmp_file = tmp_path / "gmf.json"
    mf.save_json(tmp_file)
    mf = MetaFrame.load_json(tmp_file)

    df_syn = mf.synthesize(100)
    df_syn_1 = mf.synthesize(100, seed=1234)
    df_syn_2 = mf.synthesize(100, seed=1234)
    for col in df_syn.columns:
        assert all(df_syn_1[col].drop_nulls() == df_syn_2[col].drop_nulls())

    for col, dtype in demo_class.schema.items():
        assert dtype == df_syn[col].dtype


def test_demo_non_exist():
    """Check that trying to get a demo dataset that doesn't exist raises an error."""
    with pytest.raises(ValueError):
        demo_file("non-existing-dataset")


def test_create_dataset(tmpdir):
    """Test re-creation of the test dataset."""
    test_class = _get_demo_class("test")
    test_class.create(tmpdir/"test.csv")
    df = pl.read_csv(Path(tmpdir/"test.csv"))
    assert isinstance(df, pl.DataFrame)



"""Module with privacy classes to be used for creating GMF files."""
from __future__ import annotations

from abc import ABC, abstractmethod
from importlib.metadata import entry_points
from typing import TYPE_CHECKING, Optional, Type, Union

from metasyn.util import get_registry

if TYPE_CHECKING:
    from metasyn.distribution.base import BaseFitter


class BasePrivacy(ABC):
    """Abstract base class for privacy levels.

    This class serves as a blueprint for privacy classes. Derived classes
    should at least set the class variable `name` and implement the `to_dict`
    method, which should return a dictionary that gives the privacy type and
    its parameters.
    """

    name = "unknown_privacy"

    def __init__(self):
        pass

    @abstractmethod
    def to_dict(self) -> dict:
        """Create a dictionary that gives the privacy type, and parameters."""
        return {
            "name": self.name,
            "parameters": {},
        }

    def is_compatible(self, dist: Union[BaseFitter, Type[BaseFitter]]) -> bool:
        """Check whether the distribution has the same privacy.

        Arguments
        ---------
        dist:
            Distribution to check.
        """
        return dist.privacy_type == self.name

    def comment(self, var):  # noqa
        """Comment on the privacy features for the TOML GMF file.

        Parameters
        ----------
        col_name
            Name of the column.

        Returns
        -------
            A comment on the privacy features.
        """
        return "Above are the parameters for the column {var.name}"


class BasicPrivacy(BasePrivacy):
    """Class representing no privacy level.

    This class uses statistically optimal distributions. It inherits from the
    `BasePrivacy` class and sets the `name` attribute to "none".
    """

    name = "none"

    def to_dict(self) -> dict:
        return BasePrivacy.to_dict(self)


def get_privacy(name: str, parameters: Optional[dict] = None) -> BasePrivacy:
    """Create a new privacy object using a name and parameters.

    Parameters
    ----------
    name
        Name of the privacy type, use "none" for no specific type of privacy.
    parameters
        The parameters for the privacy type. This could be the epsilon for differential
        privacy or partition_size for disclosure control, by default None.

    Returns
    -------
        A new instantiated object for privacy.

    Raises
    ------
    KeyError
        If the name of the privacy type cannot be found.
    """
    parameters = parameters if parameters is not None else {}
    for entry in entry_points(group="metasyn.privacy"):
        if name == entry.name:
            return entry.load()(**parameters)

    # Handle case where the plugin is not installed or is misspelled.
    registry = get_registry()
    available_plugins = {key: val for key, val in registry.items() if name in val["privacy"]}
    if len(available_plugins) > 0:
        avail = "\n".join(f"{key}: {val['url']}" for key, val in available_plugins.items())
        raise ImportError(f"No plugin is installed that provides '{name}' privacy. "
                          f"Available plugins that provide '{name}' privacy:\n\n{avail}")
    privacy_names = [entry.name for entry in entry_points(group="metasyn.privacy")]
    avail = "\n".join(f"<{key}> provides: {val['privacy']}\n\t{val['url']}"
                      for key, val in registry.items() if len(val['privacy']) > 0)
    raise ImportError(f"Unknown privacy type with name '{name}'. "
                       "Ensure that you have installed the correct privacy package"
                       " (and not misspelled it).\n"
                      f"Installed privacy types: {privacy_names}.\n\n"
                      f"List of plugins that provide privacy:\n\n{avail}\n")



"""Regex distributions and fitters."""
from typing import Optional, Union

# from lingua._constant import LETTERS, PUNCTUATION
from regexmodel import NotFittedError, RegexModel

from metasyn.distribution.base import (
    BaseDistribution,
    BaseFitter,
    UniqueDistributionMixin,
    builtin_fitter,
    metadist,
)


@metadist(name="core.regex", var_type="string", version="2.0")
class RegexDistribution(BaseDistribution):
    """Structured string distribution using regex.

    Main implementation details in the regexmodel package:
    https://github.com/sodascience/regexmodel

    This distribution tries to create a regex that fits the variable. This
    regex also contains statistical information about the probabilities if the
    regex has multiple options (e.g. a|b). The regex is only a subset what is provided
    by the python re package. What is currently implemented:

    - Parentheses with multiple options and no modifiers, e.g. ([a]|[b]|[c])
    - Square brackets without negation, e.g. [abc]
    - Ranges [A-Z], [a-z], [0-9], but not subranges (e.g. [0-3])
    - Repetition quantifiers (curly brackets) with minimum and maximum [A-Z]{3,6}, but not [A-Z]{6}.

    When fitting the RegexDistribution using the fit method, pay attention to the
    count_thres and method parameters. By default these will be dynamic and take reasonable
    values for the input, but in some cases it can be important to set them manually.
    The count_thres parameter sets the minimum number of times a regex element needs to
    be used. So, if count_thres=2, and there is only one value starting with "a", then the regex
    will never start with "a". In effect, a higher value will provide more privacy, less
    utility and a faster fit. The other parameter "method" has a small effect on the accuracy
    of the regex, and a larger effect on the worst case time consumption for fitting.
    Set to "accurate" for the best result, and "fast" for the fastest result.

    Examples that this distribution should work reasonably for are: email,
    ID's, telephone numbers, ip addresses, etc.

    Parameters
    ----------
    regex_data:
        Valid inputs for the regex model are:
        - str: String with a regex (that falls within the specifications, see above).
        - dict: Serialized version of the regex model, as it is coming from a JSON file.
        - RegexModel: Initialized regex model.

    Examples
    --------
    >>> RegexDistribution(r"AB[0-9]{4}").draw()
    "AB8123"
    >>> RegexDistribution(r"(a|b|c)10)").draw()
    "b10"
    """

    def __init__(self, regex_data: Union[str, dict, RegexModel]):
        self.regex_model = RegexModel(regex_data)

    def draw(self):
        return self.regex_model.draw()

    def _param_dict(self):
        return {"regex_data": self.regex_model.serialize()}

    @property
    def _params_formatted(self):
        return f"\t- regex: {self.regex_model.regex}"

    @classmethod
    def _param_schema(cls):
        return {
            "regex": {
                "type": "string"
            },
            "counts": {
                "type": "array"
            }
        }

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls(r"[ABC][0-9]{3,4}")

    def information_criterion(self, values):
        mean_len = values.str.len_chars().mean()
        diff = ((values.str.len_chars() - mean_len)**2).mean()
        if mean_len > 3 and diff < 1:
            return -2
        return 0


@builtin_fitter(distribution=RegexDistribution, var_type="string", version="2.0")
class RegexFitter(BaseFitter):
    """Fitter for regex distribution."""

    distribution: type[RegexDistribution]

    def _fit(self, series, count_thres: Optional[int] = None, method: str = "auto"):
        """Fit a regex to structured strings.

        Arguments
        ---------
        series:
            Values to be fitted (pl.Series).
        count_thres:
            Threshold for regex elements, so that a regex element can only be used if
            the number of series satisfying said element is higher than the threshold.
        method:
            Method for fitting the regex model. Possible series are ["accurate", "fast", "auto"]
            The "auto" method switches between the "accurate" and "fast" methods depending on
            the number of characters (fast if #char > 10000) in the series.
        """
        if method == "auto":
            if series.str.len_chars().mean() > 10:
                method = "fast"
            else:
                method = "accurate"

        # Make count_thres ~= #series/100 up to 50 if in auto mode.
        if count_thres is None:
            count_thres = min(50, max(2, round(len(series)/50)))

        # Try to fit the series, if it cannot be fit, then use the default distribution.
        try:
            model = RegexModel.fit(series, count_thres=count_thres, method=method)
        except NotFittedError:
            return self.distribution.default_distribution()
        return self.distribution(model)



@metadist(name="core.regex", var_type="string", unique=True)
class UniqueRegexDistribution(UniqueDistributionMixin, RegexDistribution):
    """Unique variant of the regex distribution.

    See :class:`~RegexDistribution` for examples and explanation on this distribution.
    """

@builtin_fitter(distribution=UniqueRegexDistribution, var_type="string")
class UniqueRegexFitter(RegexFitter):
    """Fitter for unique regex distribution."""



"""Utility module for metasyn."""
from __future__ import annotations

import random

import faker
import numpy as np
import polars as pl

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # type: ignore  # noqa

from importlib.resources import files

ALL_VAR_TYPES = ["discrete", "continuous", "time", "date", "datetime", "string", "categorical"]

def get_registry() -> dict:
    """Get the registry dictionary from the package.

    This registry contains information on plugins that are available for metasyn.

    Returns
    -------
        Dictionary containing the registry entries.
    """
    registry_fp = files(__package__) / "schema" / "plugin_registry.toml"
    with registry_fp.open("rb") as handle:
        registry = tomllib.load(handle)
    return registry

def set_global_seeds(seed: int):
    """Set the global seeds for all random number generators.

    Parameters
    ----------
    seed
        The seed to use for the random number generators.
    """
    np.random.seed(seed)
    random.seed(seed)
    faker.Faker.seed(seed)


def get_var_type(series: pl.Series) -> str:
    """Convert polars dtype to metasyn variable type.

    Parameters
    ----------
    series:
        Series to get the metasyn variable type for.

    Returns
    -------
    var_type:
        The variable type that is found.
    """
    if not isinstance(series, pl.Series):
        series = pl.Series(series)
    polars_dtype = str(series.dtype.base_type())

    convert_dict = {
        "Int8": "discrete",
        "Int16": "discrete",
        "Int32": "discrete",
        "Int64": "discrete",
        "UInt8": "discrete",
        "UInt16": "discrete",
        "UInt32": "discrete",
        "UInt64": "discrete",
        "Float32": "continuous",
        "Float64": "continuous",
        "Date": "date",
        "Datetime": "datetime",
        "Time": "time",
        "String": "string",
        "Categorical": "categorical",
        "Enum": "categorical",
        "Boolean": "categorical",
        "Null": "continuous",
    }
    try:
        return convert_dict[polars_dtype]
    except KeyError as exc:
        raise TypeError(f"Unsupported polars type '{polars_dtype}'") from exc



"""Create and retrieve demo datasets."""

# import random
import string
from abc import ABC, abstractmethod
from datetime import date, datetime, time, timedelta
from importlib.resources import files
from pathlib import Path

import numpy as np
import polars as pl

from metasyn.varspec import VarSpec

_AVAILABLE_DATASETS = {}


def register(*args):
    """Register a dataset so that it can be found by name."""

    def _wrap(cls):
        _AVAILABLE_DATASETS[cls().name] = cls()
        return cls

    return _wrap(*args)


class BaseDataset(ABC):
    """Base class for demo datasets."""

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    def file_location(self):
        return files(__package__) / f"demo_{self.name}.csv"

    def get_dataframe(self):
        return pl.read_csv(self.file_location, schema_overrides=self.schema, try_parse_dates=True)

    @property
    @abstractmethod
    def schema(self):
        pass

    @property
    def var_specs(self):
        return []


@register
class TitanicDataset(BaseDataset):
    """Included in pandas, but post-processed to contain more columns."""

    @property
    def name(self):
        return "titanic"

    @property
    def schema(self):
        return {"Sex": pl.Categorical, "Embarked": pl.Categorical}

    @property
    def var_specs(self):
        return [VarSpec("PassengerId", unique=True)]


@register
class SpaceShipDataset(BaseDataset):
    """CC-BY from https://www.kaggle.com/competitions/spaceship-titanic."""

    @property
    def name(self):
        return "spaceship"

    @property
    def schema(self):
        return {
            "HomePlanet": pl.Categorical,
            "CryoSleep": pl.Categorical,
            "VIP": pl.Categorical,
            "Destination": pl.Categorical,
            "Transported": pl.Categorical,
        }


@register
class FruitDataset(BaseDataset):
    """Very basic example data from Polars."""

    @property
    def name(self):
        return "fruit"

    @property
    def schema(self):
        return {"fruits": pl.Categorical, "cars": pl.Categorical}

    @property
    def var_specs(self):
        return [VarSpec("ID", unique=True), VarSpec("B", unique=False)]


@register
class SurveyDataset(BaseDataset):
    """Columns from ESS round 11 Human Values Scale questionnaire for the Netherlands."""

    @property
    def name(self):
        return "survey"

    @property
    def schema(self):
        return {}


@register
class SyntheaImagingDataset(BaseDataset):
    """Synthetic medical health dataset from Synthea.

    Jason Walonoski, Mark Kramer, Joseph Nichols, Andre Quina, Chris Moesel, Dylan Hall,
    Carlton Duffett, Kudakwashe Dube, Thomas Gallagher, Scott McLachlan,
    Synthea: An approach, method, and software mechanism for generating synthetic patients
    and the synthetic electronic health care record, Journal of the American Medical Informatics
    Association, Volume 25, Issue 3, March 2018, Pages 230–238, https://doi.org/10.1093/jamia/ocx079
    """

    @property
    def name(self):
        return "synthea_imaging"

    @property
    def schema(self):
        return {"SOP_DESCRIPTION": pl.Categorical, "BODYSITE_DESCRIPTION": pl.Categorical}


@register
class HospitalDataset(BaseDataset):
    """Example electronic health record hospital dataset.

    This dataset was created manually by the metasyn team.
    """

    @property
    def name(self):
        return "hospital"

    @property
    def schema(self):
        return {"date_admitted": pl.Date, "time_admitted": pl.Time, "type": pl.Categorical}


@register
class DrugUseDataset(BaseDataset):
    """Example dataset with answers to an open question on study participants' daily drug use.

    This example dataset was generated through ChatGPT-4o on 07-11-2024 using the following prompt:
    > Create a csv with 12 rows and 2 columns: participant_id, and drug_use. The participant_id
    has a standard alphanumeric structure, and the drug_use contains participant's responses on
    how they use drugs in their daily life.
    """

    @property
    def name(self):
        return "druguse"

    @property
    def schema(self):
        return {}


@register
class TestDataset(BaseDataset):
    """Test dataset with all supported data types."""

    @property
    def name(self):
        return "test"

    @property
    def schema(self):
        columns = pl.read_csv(self.file_location).columns
        return {
            col_name: (getattr(pl, col_name) if col_name != "NA" else pl.String)
            for col_name in columns
        }

    @classmethod
    def create(cls, csv_file):
        all_series = []
        n_rows = 100

        for int_val in [8, 16, 32, 64]:
            all_series.append(
                pl.Series(
                    f"Int{int_val}",
                    [np.random.randint(-10, 10) for _ in range(n_rows)],
                    dtype=getattr(pl, f"Int{int_val}"),
                )
            )
            all_series.append(
                pl.Series(
                    f"UInt{int_val}",
                    [np.random.randint(10) for _ in range(n_rows)],
                    dtype=getattr(pl, f"UInt{int_val}"),
                )
            )

        for float_val in [32, 64]:
            all_series.append(
                pl.Series(
                    f"Float{float_val}",
                    np.random.randn(n_rows),
                    dtype=getattr(pl, f"Float{float_val}"),
                )
            )

        all_series.append(
            pl.Series(
                "Date",
                [date(2024, 9, 4) + timedelta(days=i) for i in range(n_rows)],
                dtype=pl.Date,
            )
        )
        all_series.append(
            pl.Series(
                "Datetime",
                [
                    datetime(2024, 9, 4, 12, 30, 12)
                    + timedelta(hours=i, minutes=i * 2, seconds=i * 3)
                    for i in range(n_rows)
                ],
                dtype=pl.Datetime,
            )
        )
        all_series.append(
            pl.Series(
                "Time",
                [time(3 + i // 20, 6 + i // 12, 12 + i // 35) for i in range(n_rows)],
                dtype=pl.Time,
            )
        )
        all_series.append(
            pl.Series(
                "String", np.random.choice(list(string.printable), size=n_rows), dtype=pl.String
            )
        )
        all_series.append(
            pl.Series(
                "Utf8", np.random.choice(list(string.printable), size=n_rows), dtype=pl.Utf8
            )
        )
        all_series.append(
            pl.Series(
                "Categorical",
                np.random.choice(list(string.ascii_uppercase[:5]), size=n_rows),
                dtype=pl.Categorical,
            )
        )
        all_series.append(
            pl.Series("Boolean", np.random.choice([True, False], size=n_rows), dtype=pl.Boolean)
        )
        all_series.append(pl.Series("NA", [None for _ in range(n_rows)], dtype=pl.String))

        # Add NA's for all series except the categorical
        for series in all_series:
            if series.name != "Categorical":
                none_idx = np.random.choice(np.arange(n_rows), size=n_rows // 10, replace=False)
                none_idx.sort()
                series[none_idx] = None

        # Write to a csv file
        pl.DataFrame(all_series).write_csv(csv_file)


def _get_demo_class(name):
    if name in _AVAILABLE_DATASETS:
        return _AVAILABLE_DATASETS[name]
    raise ValueError(
        f"No demonstration dataset with name '{name}'. Options: {list(_AVAILABLE_DATASETS)}."
    )


def demo_file(name: str = "titanic") -> Path:
    """Get the path for a demo data file.

    There are six options:
        - titanic (Included in pandas, but post-processed to contain more columns)
        - spaceship (CC-BY from https://www.kaggle.com/competitions/spaceship-titanic)
        - synthea_imaging (CC-BY from https://synthea.mitre.org/downloads)
        - fruit (very basic example data from Polars)
        - survey (columns from ESS round 11 Human Values Scale questionnaire for the Netherlands)
        - test (columns with all supported data types)

    Arguments
    ---------
    name:
        Name of the demo dataset.

    Returns
    -------
        Path to the dataset.

    References
    ----------
    European Social Survey European Research Infrastructure (ESS ERIC). (2024). ESS11 integrated
    file, edition 1.0 [Data set]. Sikt - Norwegian Agency for Shared Services in Education and
    Research. https://doi.org/10.21338/ess11e01_0
    """
    return _get_demo_class(name).file_location


def demo_dataframe(name: str = "titanic") -> pl.DataFrame:
    """Get a demonstration dataset as a prepared polars dataframe.

    There are six options:
        - titanic (Included in pandas, but post-processed to contain more columns)
        - spaceship (CC-BY from https://www.kaggle.com/competitions/spaceship-titanic)
        - synthea_imaging (CC-BY from https://synthea.mitre.org/downloads)
        - fruit (very basic example data from Polars)
        - survey (columns from ESS round 11 Human Values Scale questionnaire for the Netherlands)
        - test (columns with all supported data types)

    Arguments
    ---------
    name:
        Name of the demo dataset: spaceship, fruit, or titanic.

    Returns
    -------
        Polars dataframe with correct column types

    References
    ----------
    European Social Survey European Research Infrastructure (ESS ERIC). (2024). ESS11 integrated
    file, edition 1.0 [Data set]. Sikt - Norwegian Agency for Shared Services in Education and
    Research. https://doi.org/10.21338/ess11e01_0
    """
    return _get_demo_class(name).get_dataframe()



"""Tests I/O for toml configuration and GMF files."""
from filecmp import cmp
from pathlib import Path

import pytest

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from pytest import mark

from metasyn.distribution import builtin_fitters
from metasyn.metaframe import MetaFrame
from metasyn.testutils import create_input_toml, create_md_report


def test_datafree_create(tmpdir):
    """Test whether a synthetic dataset can be generated without data."""
    temp_toml = tmpdir / "test.toml"
    create_input_toml(temp_toml)
    assert cmp(temp_toml, Path("examples", "config_files", "example_all.toml"))
    mf = MetaFrame.fit_dataframe(None, config=Path(temp_toml))

    assert isinstance(mf, MetaFrame)
    assert mf.n_columns == len(builtin_fitters)

@mark.parametrize(
    "toml_input,data", [
        (Path("examples", "config_files", "example_all.toml"), None),
        # (Path("examples", "config_files", "example_config.toml"), demo_dataframe("titanic")),
        (Path("examples", "config_files", "example_datafree.toml"), None),
    ]
)
def test_toml_save_load(tmpdir, toml_input, data):
    """Test whether TOML GMF files can be saved/loaded."""
    mf = MetaFrame.fit_dataframe(data, config=toml_input)
    mf.save(tmpdir/"test.toml")
    new_mf = MetaFrame.load(tmpdir/"test.toml")
    assert mf.n_columns == new_mf.n_columns

def test_varspec_update():
    """Check whether overwriting the  varspecs with the var_specs parameter works."""
    toml_input = Path("examples", "config_files", "example_all.toml")
    var_specs = [{
        "name": "DiscreteTruncatedNormalDistribution",
        "var_type": "discrete",
        "distribution": {
            "name": "core.normal",
            "unique": False,
            "parameters": {
                "mean": 0,
                "sd": 1,
            }
        }
    }]
    mf_normal = MetaFrame.fit_dataframe(None, config=toml_input)
    mf_varspec = MetaFrame.fit_dataframe(None, var_specs=var_specs, config=toml_input)
    assert mf_normal["DiscreteTruncatedNormalDistribution"].distribution.name == "core.truncated_normal"
    assert mf_varspec["DiscreteTruncatedNormalDistribution"].distribution.name == "core.normal"

@mark.parametrize(
    "gmf_file", [
        Path("examples", "gmf_files", "example_gmf_simple.json"),
        Path("examples", "gmf_files", "example_gmf_titanic.json"),
    ]
)
def test_md_output(tmpdir, gmf_file):
    """Test whether the markdown report can be generated from the GMF file."""
    create_md_report(gmf_file, tmpdir/"test.md")
    assert Path(tmpdir/"test.md").is_file()
    assert isinstance(MetaFrame.load(gmf_file), MetaFrame)

def test_toml_err():
    """Test whether errors are correctly handled."""
    with pytest.raises(FileNotFoundError):
        MetaFrame.from_config("doesnotexist.toml")

    with pytest.raises(tomllib.TOMLDecodeError):
        MetaFrame.from_config(Path("tests", "data", "bad_config.toml"))

    with pytest.raises(ValueError):
        MetaFrame.from_config(Path("tests", "data", "unsupported_config.toml"))

    with pytest.raises(ValueError):
        MetaFrame.from_config(Path("tests", "data", "incompatible_config.toml"))



"""Module implementing the Poisson distribution."""

import numpy as np
from scipy.stats import poisson

from metasyn.distribution.base import (
    BaseDistribution,
    BaseFitter,
    ScipyDistribution,
    builtin_fitter,
    metadist,
)


@metadist(name="core.poisson", var_type="discrete")
class PoissonDistribution(ScipyDistribution):
    """Poisson distribution.

    Parameters
    ----------
    rate:
        Rate (mean) of the poisson distribution.

    Examples
    --------
    >>> PoissonDistribution(rate=3.5)
    """

    dist_class = poisson

    def __init__(self, rate: float):
        self.par = {"rate": rate}
        self.dist = self.dist_class(mu=rate)

    def _information_criterion(self, values):
        return np.log(len(values))*self.n_par - 2*np.sum(self.dist.logpmf(values))

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls(0.5)

    @classmethod
    def _param_schema(cls):
        return {
            "rate": {"type": "number"},
        }


@builtin_fitter(distribution=PoissonDistribution, var_type="discrete")
class PoissonFitter(BaseFitter):
    """Fitter for Poisson distribution."""

    def _fit(self, series):
        mean_series = series.mean()
        mean_series = mean_series if mean_series >= 0 else 0
        return self.distribution(mean_series)
